from typing import Callable, TypeVar
from requests import Response
import json


T = TypeVar('T')
U = TypeVar('U')
def _tryParse(constr: Callable[[T], U], data: dict[str, T], key: str, default: U) -> U:
    val = data.get(key)
    if val:
        return constr(val)
    else:
        return default

class ResponseModel:
    def __init__(self, res: Response) -> None:
        self.Code = res.status_code
        self.data: dict = json.loads(res.text) if self.Code==200 else {}


from django.http import HttpResponse, JsonResponse
from django.core.handlers.wsgi import HttpRequest
from django.http.request import QueryDict
from django.middleware.csrf import get_token
from django.views.decorators.http import require_http_methods

from limes_common.models.network import server
from limes_common import config
from . import fileHandler


def _toDict(qd: QueryDict):
    d = {}
    for k, v in qd.items():
        d[k] = v
    return d

# doesn't work
def _getClientIp(request: HttpRequest) -> str:
    x_forwarded_for = str(request.META.get('HTTP_X_FORWARDED_FOR'))
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    elif request.META.get('HTTP_X_REAL_IP'):
        ip = str(request.META.get('HTTP_X_REAL_IP'))
    else:
        ip = str(request.META.get('REMOTE_ADDR'))
    return ip

class Client:
    # todo: add timeout
    def __init__(self, res: server.Login.Request) -> None:
        self.Token = res.Token
        self.FirstName = res.FirstName
        self.LastName = res.LastName

_activeClients: dict[str, Client] = {}
_clientsByToken: dict[str, str] = {} # token: clientId

# these must match those in limes_common.models.network.endpoints

@require_http_methods(['GET'])
def Init(request: HttpRequest):
    return JsonResponse({config.CSRF_KEY: get_token(request)})

@require_http_methods(['POST'])
def Login(request: HttpRequest):
    SL = server.Login
    res = SL.Request(_toDict(request.POST))

    # remove old if exists
    old = _clientsByToken.get(res.Token)
    if old is not None: 
        _activeClients.pop(old)
        _clientsByToken.pop(res.Token)

    _activeClients[res.Id] = Client(res)
    _clientsByToken[res.Token] = res.Id

    print('login: %s' % (res.FirstName))
    return JsonResponse(SL.MakeResponse(True))

@require_http_methods(['POST'])
def Authenticate(request: HttpRequest):
    SA = server.Authenticate
    res = SA.Request(_toDict(request.POST))
    client = _activeClients.get(res.Id)
    success = client is not None
    token = client.Token if success else ''
    fName = client.FirstName if success else ''
    lName = client.LastName if success else ''
    print('auth: %s' % (fName if success else 'unknown'))
    return JsonResponse(SA.MakeResponse(success, token, fName, lName))

@require_http_methods(['POST'])
def Add(request: HttpRequest):
    SA = server.Add
    fail = lambda m='': JsonResponse(SA.MakeResponse(False, message=m))

    print('adding...')

    req = SA.Request(_toDict(request.POST))
    client = _activeClients.get(req.ClientId)
    if not client: return fail('Not logged in')

    success, msg = fileHandler.TryAddFile(client.Token, req.Meta, request.FILES[SA.FILE_KEY])
    if success:
        return JsonResponse(SA.MakeResponse(True, sampleName=msg))
    else:
        return fail(msg)

@require_http_methods(['POST'])
def Blast(request: HttpRequest):
    SB = server.Blast
    result = fileHandler.Blast(request.FILES[SB.FILE_KEY])
    return JsonResponse(SB.MakeResponse(True, result))



class Prod():
    VERIFY_CERTIFICATE = True
    KEEP_RAW_DATA = False
    ENCODING = 'utf-8'

class Dev(Prod):
    # VERIFY_CERTIFICATE = False
    KEEP_RAW_DATA = True

class ActiveGeneric(Dev):
    pass
# ---------------------------------------

class ProdCore(Prod):
    URL = '' # todo
    PORT = 0 # todo

class DevCore(ProdCore, Dev):
    URL = 'localhost'
    PORT = 3388

class ActiveCore(DevCore):
    pass
# ---------------------------------------


class ProdClient(Prod):
    CORE_ADDRESS = ProdCore.URL
    ELAB_URL = 'https://us.elabjournal.com/api/v1/'
    ELAB_TIME_FORMAT = '%Y-%m-%dT%H:%M:%SZ'

class DevClient(ProdClient, Dev):
    CORE_ADDRESS = '%s:%s/' % ('https://127.0.0.1', DevCore.PORT)
    CREDENTIALS_PATH = 'credentials' # todo: make more secure

class ActiveClient(DevClient):
    pass
# ---------------------------------------

class ProdProvider(Prod):
    CORE_ADDRESS = ProdCore.URL
    URL = '127.0.0.1'
    PORT = 3389

class DevProvider(ProdProvider, Dev):
    CORE_ADDRESS = DevClient.CORE_ADDRESS
    pass

class ActiveProvider(DevProvider):
    pass


CSRF_KEY = 'csrfmiddlewaretoken'
ELAB_URL = 'https://us.elabjournal.com/'
# ELAB_URL = 'https://elab.msl.ubc.ca/'
ELAB_API = '%sapi/v1/' % ELAB_URL

SERVER_API_VER = 'd1'
SERVER_URL = 'http://127.0.0.1:8001/api/%s/' % SERVER_API_VER
SERVER_DB_PATH = 'limes_server/db'

SHAMWOW_URL = 'shamwow.microbiology.ubc.ca'
            


from . import limes

import sys
from typing import List
from getpass import getpass
from functools import reduce

from .tools.qol import Switch, T
from .res import strings as Constants
from limes_common import config

def main(args:List[str] = sys.argv[1:]):
    # for later abstraction of help
    class Command:
        def __init__(self, schemas, options) -> None:
            self.schemas = schemas
            self.options = options

        def printHelp(self):
            self.printLine()
            print('Required:')

        def printLine(self):
            print('====================================================================')

    def Login():
        userKeys = ['-u', '--username']
        passKeys = ['-p', '--password']

        tries = 3
        user = input('Username: ')
        for _ in range(tries):
            try:
                password = getpass()
            except KeyboardInterrupt:
                return

            if limes.Login(user, password):
                return
        return '%s failed attempts, check your connection and credentials' % tries

    def Add():
        if len(args) == 5:
            limes.Add(args[2], args[1], args[4])
        elif len(args) == 3:
            limes.Add(args[2], args[1])
        else:
            return 'usage: limes add [file path] [sampleId] (-as [name])'
        return ''

    def Search():
        try:
            if not limes._auth(): return
        except:
            return

        if len(args) < 2:
            return 'usage: Search [tokens], where [tokens] is any string'

        query = ""
        for arg in args[1:]:
            query += arg
        samples = limes.Search(query)
        count = len(samples)
        msg = '%s result%s found' % (count, 's' if count != 1 else '')
        max = 3
        if count > max: msg += ', truncating to first %s' % max
        i = 1
        for sample in samples:
            msg += '\n'
            # msg += '\n%s of %s' % (i, count)
            msg += '\nID: %s\nName: %s' % (sample.Id, sample.Name)
            msg += '\n%smembers/inventory/browser/?sampleID=%s#viewSample' % (config.ELAB_URL, sample.Id)
            i += 1
            if i > max: break
        return msg

    def Blast():
        try:
            if not limes._auth(): return
        except:
            return

        if len(args) != 2:
            return 'usage: Blast [query], where [query] is a valid .fasta for blast queries'

        return limes.Blast(args[1])

    def _printHelp(unkownArg=False):
        msg = 'use [-h] or [--help] to view options\nor view available documentation at https://github.com/Tony-xy-Liu/Limes/'
        if unkownArg:
            msg = '[%s] was not recognized\n%s' % (unkownArg, msg)
        return msg

    def _getVal(keys: List[str], args: List[str]) -> tuple[bool, str]:
        next = 1
        for item in args:
            if item in keys and next < len(args):
                return True, args[next]
            next +=1
        return False, ""

    arg = args[0] if len(args) > 0 else ''

    def base():
        print(Constants.version)
        limes._auth()

    # the switch doesn't need to be fast, make it easier to make
    # also need to add hints here
    return Switch(arg, {
        '': base,
        '-h': lambda: Constants.help,
        '--help': lambda: Constants.help,
        'login': Login,
        'dlogin': lambda: limes.dLogin(),
        'search': Search,
        'blast': Blast,
        'add': Add,
        'test': lambda: limes.Test(),
    }, lambda: _printHelp(arg))

if __name__ == "__main__":
    end = main()
    if not end is None and end != '':
        sys.exit(end)
    else:
        sys.exit()


from io import BufferedReader
from typing import Any, Union
import os
from functools import reduce
from requests.exceptions import ConnectionError

from paramiko import file
from limes.tools.qol import Switch

from limes_common.connections import ELabConnection, ServerConnection
from limes_common.models.network import server
from limes_common.models.network.elab import SampleModel

# from  import Sample
# from models.network import HttpMethod
# from coms.requester import Requester
# from requests import Request as py_Requester


# class Query():
#     def ByAll(searchString: str) -> Tuple[bool, List[Sample]]:
#         ENDPOINT = 'samples'
#         endpointWithParam = '%s?search=%s' % (ENDPOINT, searchString)
#         succeed, response = _sendAuthenticatedRequest(
#             Config.ELAB_URL, endpointWithParam, HttpMethod.GET)
#         if succeed:
#             count = response['recordCount']
#             total = response['totalRecords']
#             if total > count:
#                 print('too many hits (%s total), truncated to %s results' %
#                     (total, count))
#             return True, list(map(lambda x: Sample(x, Config.ELAB_TIME_FORMAT), response['data']))
#         else:
#             return False, None
_server = ServerConnection()
_eLab = ELabConnection()

# def _auth() -> tuple[bool, str]:
def _auth() -> bool:
    res = _server.Authenticate()
    if res.Success:
        print('Authenticated terminal as %s' % res.FirstName)
        # return True, res.FirstName
        _eLab.SetToken(res.ELabKey)
    else:
        # return False, ''
        print('Not logged in')
    return res.Success
            

def Login(username, password) -> bool:
    if not _server.Ready: return False
    res = _eLab.Login(username, password)
    if res.Code not in [200, 401]:
        raise Exception('Failed to connect to eLab')

    if res.Success:
        _server.Login(res.Token, res.FirstName, res.LastName)
        print('logged in terminal as %s' % res.FirstName)
    return res.Success

# class UnrecognizedCriteriaException(Exception):
#     pass
# SEARCH_CRITERIA_LIST = list(c.lower() for c in ['sampleID', 'name'])
# def Search(params: dict[str, list[str]]):
#     params = dict((k.lower(), v) for k, v in params.items())
#     invalids = list(filter(lambda p: p not in SEARCH_CRITERIA_LIST, params.keys()))
#     if len(invalids) > 0:
#         raise UnrecognizedCriteriaException('unrecognized criteria: %s' % \
#             reduce(lambda s, c: '%s, [%s]' % (s, c), invalids[1:], '[%s]' %invalids[0]))

#     switcher = {
#         'sampleid': _eLab.SearchSamplesById
#     }

#     # todo, this should go in main, return dict here
#     # todo value check
#     try:
#         res = list(switcher[k](v) for k, v in params.items())
#     except ValueError as err:
#         return err
#     # todo: AND and OR logic
#     # todo: save result as csv
#     res = res[0] # temp since only one criteria so far
#     count = len(res.Samples)
#     msg = '%s result%s found' % (count, 's' if count != 1 else '')
#     max = 5
#     i = 1
#     for sample in res.Samples:
#         msg += '\n\n%s of %s' % (i, count)
#         msg += '\nID: %s\nName: %s' % (sample.Id, sample.Name)
#         i += 1
#         if i > max: break
#     return msg

def Search(token: str) -> list[SampleModel]:
    res = _eLab.SearchSamples(token)
    return res.Samples

def Add(sampleId: str, path: str, fileName: str=None) -> bool:
    if not _server.Ready: return False

    if fileName is None:
        tokens = path.split('/')
        fileName = tokens[len(tokens) - 1]

    path = os.path.abspath(path)
    file = open(path, 'rb')
    res = _server.Add(sampleId, path, fileName, file)

    if not res.Success:
        print(res.Message)
    else:
        print('file [%s] added to sample [%s]' % (fileName, res.SampleName))
    return res.Success

def Blast(queryPath: str) -> str:
    try:
        query = open(queryPath, 'rb')
        res = _server.Blast(query)
        return res.Result
    except:
        print('[%s] not found' % (queryPath))
    return ''

def Test() -> None:
    # from limes_common.connections import ShamwowConnection as SC
    # sc = SC()
    _auth()
    # res = _eLab.GetSampleMeta('005000013194361')
    res = _eLab.UpdateSampleMeta('005000013194361', 'Data - Fasta', '* This field is managed by Limes, please do not edit *')
    print(res.__dict__)
    # res = _eLab.SearchSamplesByName('Demo')
    # print(res.Samples)

def dLogin() -> None:
    with open('../../credentials/elab') as cred:
        lines = cred.readlines()
        lines = list(map(lambda l: l.replace('\n', ''), lines))
        if not Login(lines[0], lines[1]):
            print('quick login failed')
# def _sendAuthenticatedRequest(url: str, endpoint: str, method: HttpMethod, body: dict = None) -> Tuple[bool, dict]:
#     if _apiToken is not None:
#         headers = {
#             'Authorization': _apiToken
#         }
#         return SendRequest(url, endpoint, method, headers, str(body))
#     else:
#         print('You must login first with `Limes.Login()`')
#         return False, {}


# print(Login())
# s, q = Query.ByAll('sample 12345')
# print(s)
# print(len(q))
# print(q[0]._raw)



from limes_common.models.network.server import Login
from ..basic import AdvancedEnum, AbbreviatedEnum

class Http(AbbreviatedEnum):
    GET = 1,
    POST = 2,
    PUT = 3

class Endpoint(AdvancedEnum, AbbreviatedEnum):
    def __init__(self, _:int, methods: list[Http], path: str) -> None:
        self.ValidMethods: list[Http] = methods
        self.path = path

class ServerEndpoint(Endpoint):
    LOGIN = 1, [Http.POST], 'login'
    AUTHENTICATE = 2, [Http.POST], 'authenticate'
    INIT = 3, [Http.GET], 'init'
    ADD = 4, [Http.POST], 'add'
    BLAST = 5, [Http.POST], 'blast' # post becuase need to send file

class ELabEndpoint(Endpoint):
    LOGIN = 1, [Http.POST], 'auth/user'
    SAMPLES = 2, [Http.POST, Http.GET], 'samples'


from django.urls import path
from django.urls.resolvers import URLPattern

from limes_common.models.network.endpoints import ServerEndpoint as Endpoint
from limes_common import config
from . import views

def genPath(ep: Endpoint) -> URLPattern:
    epName = str(ep)
    return path('%s/%s' % (config.SERVER_API_VER, ep.path), views.__dict__[epName.title()], name=epName)

# urlpatterns = [
#     path('%s/login' % VER, views.Login, name='login'),
#     path('%s/authenticate' % VER, views.Authenticate, name='authenticate'),
# ]

urlpatterns = list(map(genPath, Endpoint))


from io import BufferedReader
import json
from typing import Any, overload
from requests import Response as py_Response

from limes_common import config
from limes_common.models.basic import AbbreviatedEnum
from . import _tryParse, ResponseModel

class Locations(AbbreviatedEnum):
    ELAB = 1
    SHAMWOW = 2

CLIENT_ID_KEY = 'id'

class Init:
    @classmethod
    def MakeResponse(cls, csrfToken: str):
        return {config.CSRF_KEY: csrfToken}

    class Response(ResponseModel):
        def __init__(self, res: py_Response) -> None:
            super().__init__(res)
            if self.Code == 200:
                try:
                    self.Csrf = json.loads(res.text)[config.CSRF_KEY]
                except:
                    self.Csrf = ''

class Authenticate:
    @classmethod
    def MakeRequest(cls, id: str):
        return {
            CLIENT_ID_KEY: id,
            }

    @classmethod
    def MakeResponse(cls, success: bool, token: str, firstName: str, lastName: str):
        return {
            'fName': firstName,
            'lName': lastName,
            'success': success,
            'token': token,
            }

    class Request:
        def __init__(self, raw: dict) -> None:
            self.Id: str = str(raw.get('id'))

    class Response(ResponseModel):
        def __init__(self, res: py_Response) -> None:
            super().__init__(res)
            data = json.loads(res.text) if self.Code==200 else {}
            self.Success = _tryParse(bool, data, 'success', False)
            self.ELabKey = _tryParse(str, data, 'token', '')
            self.FirstName = _tryParse(str, data, 'fName', '')
            self.LastName = _tryParse(str, data, 'lName', '')

class Login:
    @classmethod
    def MakeRequest(cls, eLabKey: str, firstName: str, lastName: str):
        return {
            'fName': firstName,
            'lName': lastName,
            'token': eLabKey,
            }

    @classmethod
    def MakeResponse(cls, success: bool):
        return {
            'success': success,
            }

    class Request:
        def __init__(self, raw: dict) -> None:
            self.Token = str(raw.get('token'))
            self.Id = str(raw.get(CLIENT_ID_KEY))
            self.FirstName = str(raw.get('fName'))
            self.LastName = str(raw.get('lName'))

    class Response(ResponseModel):
        def __init__(self, res: py_Response) -> None:
            super().__init__(res)
            data = json.loads(res.text) if self.Code==200 else {}
            self.Success = _tryParse(bool, data, 'success', False)


class FileMeta:
    def __init__(self, raw:dict = {}) -> None:
        self.FilePath = str(raw.get('path'))
        self.SampleId = str(raw.get('sampleId'))
        self.FileName = str(raw.get('name'))

class Add:
    FILE_KEY = 'file'

    @classmethod
    def MakeRequest(cls, sampleId: str, path: str, fileName: str):
        return {
            'path': path,
            'sampleId': sampleId,
            'name': fileName
            }

    @classmethod
    def MakeResponse(cls, success: bool, sampleName: str='', message: str=''):
        return {
            'success': success,
            'msg': message,
            'sampleName': sampleName,
            }

    class Request:
        def __init__(self, raw: dict) -> None:
            self.ClientId = str(raw.get(CLIENT_ID_KEY))
            self.Meta = FileMeta(raw)

    class Response(ResponseModel):
        def __init__(self, res: py_Response) -> None:
            super().__init__(res)
            data = json.loads(res.text) if self.Code==200 else {}
            self.Success = _tryParse(bool, data, 'success', False)
            self.Message = _tryParse(str, data, 'msg', '')
            self.SampleName = _tryParse(str, data, 'sampleName', '')

class Blast:
    FILE_KEY = 'query'
    @classmethod
    def MakeRequest(cls):
        return {}

    @classmethod
    def MakeResponse(cls, success: bool, report: str):
        return {
            'success': success,
            'report': report,
            }
    class Response(ResponseModel):
        def __init__(self, res: py_Response) -> None:
            super().__init__(res)
            self.Success = _tryParse(bool, self.data, 'success', False)
            self.Result = _tryParse(str, self.data, 'report', '')


from io import BufferedReader
import requests as Requests
import uuid
from getpass import getuser
import os
from functools import reduce
import paramiko
from typing import Any

from limes_common.models.network import server, elab
from limes_common.models.network.endpoints import ELabEndpoint, ServerEndpoint, Endpoint
from . import config

class Connection:
    def __init__(self, url) -> None:
        self.session = Requests.session()
        self._URL: str = url

    def _makeUrl(self, ep: Endpoint):
        return (self._URL if self._URL.endswith('/') else self._URL[:-1]) + ep.path

class ServerConnection(Connection):
    def __init__(self) -> None:
        super().__init__(config.SERVER_URL)
        self._id = ('%012x:%s:%s' % (uuid.getnode(), getuser(), os.getppid()))
        try:
            res = server.Init.Response(self.session.get(self._makeUrl(ServerEndpoint.INIT)))
            if res.Code == 200:
                self._csrf = res.Csrf
            self.Ready = True
        except Requests.exceptions.ConnectionError:
            print('*** limes server is not reachable ***')
            self.Ready = False
    
    def _compile(self, data: dict[str, Any]) -> dict[str, Any]:
        data[config.CSRF_KEY] = self._csrf
        data[server.CLIENT_ID_KEY] = self._id
        return data

    def _notConnectedGuard(self):
        if not self.Ready: raise Requests.exceptions.ConnectionError('limes server not connected')

    def Authenticate(self) -> server.Authenticate.Response:
        self._notConnectedGuard()
        return server.Authenticate.Response(self.session.post(
            self._makeUrl(ServerEndpoint.AUTHENTICATE),
            data=self._compile(server.Authenticate.MakeRequest(self._id))
        ))

    def Login(self, eLabKey: str, firstName: str, lastName: str) -> server.Login.Response:
        self._notConnectedGuard()
        return server.Login.Response(self.session.post(
            self._makeUrl(ServerEndpoint.LOGIN),
            data=self._compile(server.Login.MakeRequest(eLabKey, firstName, lastName))
        ))

    def Add(self, sampleId: str, absPath: str, fileName: str, file: BufferedReader) -> server.Add.Response:
        self._notConnectedGuard()
        return server.Add.Response(self.session.post(
            self._makeUrl(ServerEndpoint.ADD),
            data=self._compile(server.Add.MakeRequest(sampleId, absPath, fileName)),
            files={server.Add.FILE_KEY: file},
        ))

    def Blast(self, query: BufferedReader) -> server.Blast.Response:
        self._notConnectedGuard()
        return server.Blast.Response(self.session.post(
            self._makeUrl(ServerEndpoint.BLAST),
            data=self._compile(server.Blast.MakeRequest()),
            files={server.Blast.FILE_KEY: query},
        ))

class ELabConnection(Connection):
    def __init__(self) -> None:
        super().__init__(config.ELAB_API)
        self._token: str = ''

    def _getAuthHeader(self):
        return {'authorization': self._token}

    def SetToken(self, token:str) -> None:
        self._token = token

    def LoggedIn(self) -> bool:
        return self._token != ''

    def Login(self, username: str, password: str) -> elab.Login.Response:
        return elab.Login.Response(self.session.post(
            self._makeUrl(ELabEndpoint.LOGIN),
            data=elab.Login.MakeRequest(username, password)
        ))

    def _truncateToId(self, id: str) -> str:
        return id[-9:] if len(id) > 9 else id

    def SearchSamplesById(self, strIds: list[str]) -> elab.Sample.ListResponse:
        ids = list(self._truncateToId(id) for id in strIds)
        query = '' if len(strIds)==0 else '?sampleID=' + reduce(lambda s, id: '%s,%s' % (s, id), ids, '')[1:]
        return elab.Sample.ListResponse(self.session.get(
            '%s/%s%s' % (self._makeUrl(ELabEndpoint.SAMPLES), 'get', query),
            headers=self._getAuthHeader()
        ))

    def SearchSamplesByName(self, name: str) -> elab.Sample.ListResponse:
        query = '' if len(name)==0 else '?name=' + reduce(lambda s, id: '%s,%s' % (s, id), [name], '')[1:]
        return elab.Sample.ListResponse(self.session.get(
            '%s%s' % (self._makeUrl(ELabEndpoint.SAMPLES), query),
            headers=self._getAuthHeader()
        ))

    def SearchSamples(self, token: str) -> elab.Sample.ListResponse:
        return elab.Sample.ListResponse(self.session.get(
            '%s%s%s' % (self._makeUrl(ELabEndpoint.SAMPLES), '?search=', token),
            headers=self._getAuthHeader()
        ))

    def GetSample(self, id: str) -> elab.Sample.Response:
        id = self._truncateToId(id)
        return elab.Sample.Response(self.session.get(
            '%s/%s' % (self._makeUrl(ELabEndpoint.SAMPLES), id),
            headers=self._getAuthHeader()
        ))

    def GetSampleMeta(self, id: str) -> elab.SampleMeta.Response:
        id = self._truncateToId(id)
        return elab.SampleMeta.Response(self.session.get(
            '%s/%s/meta' % (self._makeUrl(ELabEndpoint.SAMPLES), id),
            headers=self._getAuthHeader()
        ))

    def UpdateSampleMeta(self, sampleId: str, metaKey: str, newValue: str) -> elab.SampleMeta.UpdateResponse:
        id = self._truncateToId(sampleId)
        meta = self.GetSampleMeta(id)
        fieldId=0
        field = elab.MetaField()
        for k, f in meta.Fields.items():
            if k == metaKey:
                fieldId = f.sampleMetaID
                field = f
                if field.sampleDataType in ['TEXTAREA']:
                    field.value = newValue
                else:
                    print('cannot update a non text meta field')
        return elab.SampleMeta.UpdateResponse(self.session.patch(
            '%s/%s/meta/%s' % (self._makeUrl(ELabEndpoint.SAMPLES), id, fieldId),
            headers=self._getAuthHeader(),
            data=field.ToDict(),
        ))
class ShamwowConnection():
    def __init__(self) -> None:
        self._URL = config.SHAMWOW_URL
        ssh = paramiko.SSHClient()
        ssh.load_system_host_keys()
        login = {
            "username": "tliu",
            "hostname": self._URL
        }
        
        command = 'ls -lh'

        ssh.connect(**login)
        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(command)
        # print(ssh_stdin)
        print(''.join(ssh_stdout.readlines()))
        ssh.close()


import os
import subprocess
import threading
import time

from django.core.files.storage import default_storage
from django.core.files.uploadedfile import UploadedFile

from limes_common import config
from limes_common.connections import ELabConnection
from limes_common.models.network.elab import MetaField, SampleModel
from limes_common.models.network.server import FileMeta

BLAST_PATH = config.SERVER_DB_PATH + '/blast/'
DB_NAMES_FILE = config.SERVER_DB_PATH + '/dbNames'

def _fsPath(folder, file):
    return '%s/%s/%s' % (config.SERVER_DB_PATH, folder, file)

def TryAddFile(token: str, meta: FileMeta, file: UploadedFile) -> tuple[bool, str]:
    ec = ELabConnection()
    ec.SetToken(token)
    sampleRes = ec.GetSample(meta.SampleId)

    if sampleRes.Code != 200:
        return False, 'Sample [%s] not found' % (meta.SampleId)
    sample = sampleRes.Sample
    fsPath = _fsPath(sample.Id, meta.FileName)
    if default_storage.exists(fsPath): 
        return False, 'File [%s] already exists for sample [%s]' % (meta.FileName, sample.Name)

    try:
        default_storage.save(fsPath, file)
    except Exception as err:
        return False, str(err)

    sampleMeta = ec.GetSampleMeta(str(sample.Id))
    metaKey = 'Data - Fasta'
    trackedFiles = sampleMeta.Fields.get(metaKey, MetaField()).value.split('\n')
    newField = ''
    for f in trackedFiles:
        item = f.split(':')
        if item[0] != meta.FileName:
            newField += '%s\n' % f
    newField += '%s: %s' % (meta.FileName, meta.FilePath)

    ec.UpdateSampleMeta(str(sample.Id), metaKey, newField)

    thisDBName = BLAST_PATH + meta.FileName
    newNames: str = ''
    try:
        with default_storage.open(DB_NAMES_FILE, 'r') as dbNames:
            names = dbNames.readlines()
            names = list(map(lambda l: l.replace('\n', ''), names))
            for name in names:
                if name != thisDBName:
                    newNames += name + '\n'
    except:
        pass
    newNames += thisDBName
    with default_storage.open(DB_NAMES_FILE, 'w') as dbNames:
        dbNames.write(newNames)
        
    
    aliasList = '"%s"' % " ".join(list(map(lambda n: os.path.abspath(n), newNames.split('\n'))))
    # print(aliasList)
    aliasOut = BLAST_PATH + 'all'
    aliasOut = os.path.abspath(aliasOut)
    def makeDb():
        blast = subprocess.run(('makeblastdb -in %s -parse_seqids -dbtype nucl -out %s'%(fsPath, thisDBName)).split(' '))
        aliasCmd = ('blastdb_aliastool -dbtype nucl -title aggregate -out %s -dblist' % (aliasOut)).split(' ')
        aliasCmd.append(aliasList)
        # aliasCmd.append('"limes_server/db/blast/testmil.fasta limes_server/db/blast/122"')
        # time.sleep(3)
        # alias = subprocess.run(aliasCmd) # this doesn't work, for some reason
        os.system(" ".join(aliasCmd))
    thread = threading.Thread(target=makeDb)
    thread.start()

    print('added file: %s for sample: %s' % (meta.FileName, sample.Name))
    return True, sample.Name

def Blast(query: UploadedFile) -> str:
    queryPath = _fsPath('queries', query.name)
    default_storage.save(queryPath, query)

    out = queryPath + '.out'

    blast = subprocess.run([
        'blastn', '-query', queryPath,
        '-db', '%sall'%BLAST_PATH,
        '-out', out,
        '-outfmt', '7',
    ])
    # print(blast.stdout)
    # print(blast.stderr)
    with default_storage.open(out, 'r') as resultFile:
        result = resultFile.read()

    default_storage.delete(queryPath)
    default_storage.delete(out)
    return result


import json
from typing import TypeVar, Callable, Any
from requests import Response

from . import _tryParse, ResponseModel

class Login:
    @classmethod
    def MakeRequest(cls, username: str, password: str):
        return {
            'username': username,
            'password': password,
        }

    class Response(ResponseModel):
        def __init__(self, res: Response) -> None:
            super().__init__(res)
            self.Success = self.Code==200
            self.Token = _tryParse(str, self.data, 'token', '')
            userData = self.data['user'] if self.Success else {}
            self.FirstName = _tryParse(str, userData, 'firstName', '')
            self.LastName = _tryParse(str, userData, 'lastName', '')

class SampleModel:
    def __init__(self, data: dict = {}) -> None:
        self.Name = _tryParse(str, data, 'name', 'None')
        self.Owner =_tryParse(str, data, 'owner', '')
        self.TimeCreated =_tryParse(str, data, 'created', '')
        self.Id =_tryParse(int, data, 'sampleID', 0)
        self.Barcode =_tryParse(str, data, 'barcode', '')
        self.Description =_tryParse(str, data, 'description', '')
        self.ParentId =_tryParse(int, data, 'parent', 0)

class Sample:
    class ListResponse(ResponseModel):
        def __init__(self, res: Response) -> None:
            super().__init__(res)
            # self.Samples = list(map(lambda raw: SampleModel(raw), self.data))
            self.Samples = list(map(lambda raw: SampleModel(raw), self.data.get('data', [])))

    class Response(ResponseModel):
        def __init__(self, res: Response) -> None:
            super().__init__(res)
            self.Sample = SampleModel(self.data)


class MetaField:
    def __init__(self, data: dict = {}) -> None:
        T = TypeVar('T')
        def parse (constr, key):
            defaults = {
                bool: False,
                str: '',
                int: 0,
            }
            return _tryParse(constr, data, key, defaults[constr])
        self.raw = data
        self.archived: bool = parse(bool, 'archived')
        self.sampleMetaID: int = parse(int, 'sampleMetaID')
        self.sampleDataType: str = parse(str, 'sampleDataType') # todo, make enum
        self.key: str = parse(str, 'key')
        self.value: str = parse(str, 'value')
        self.sampleTypeMetaID: int = parse(int, 'sampleTypeMetaID')
        self.truncated: bool = parse(bool, 'truncated')

    def ToDict(self) -> dict:
        out = self.__dict__.copy()
        keys = list(out.keys())
        for k in keys:
            if k.startswith('_') or k in ['raw']:
                del out[k]
        return out

class SampleMeta:
    class Response(ResponseModel):
        def __init__(self, res: Response) -> None:
            super().__init__(res)
            self.Fields: dict[str, MetaField] = {}
            for raw in self.data['data']:
                self.Fields[raw['key']] = MetaField(raw)

    class UpdateResponse(ResponseModel):
        def __init__(self, res: Response) -> None:
            super().__init__(res)
            self.Success = self.Code == 204